/*------------------------------------------------------------ -*- C -*-
 *  2 Kernel Monte a.k.a. Linux loading Linux on x86
 *  Copyright (C) 2000 Erik Arjan Hendriks <hendriks@scyld.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: kmonte.c,v 1.9 2000/04/14 20:45:09 hendriks Exp $
 *--------------------------------------------------------------------*/

/* Auto-configuration stuff for things living outside the linux kernel
 * source tree. */
/* Include files, designed to support most kernel versions 2.0.0 and later. */
#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif

#include <linux/version.h>
#include <linux/module.h>
/* Older kernels do not include this automatically. */
#if LINUX_VERSION_CODE < 0x20300  &&  defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/pgtable.h>

/* Supporting SMP is complicated.  Stopping the processors doesn't
 * seem too bad.  Fooling with the APICs looks like it will be a major
 * pain unless the kernel exports a few more symbols.  */
#ifdef __SMP__
#error "2 Kernel Monte doesn't work with SMP!"
#endif

static unsigned long run_setup=0;
MODULE_AUTHOR("Erik Arjan Hendriks <hendriks@scyld.com>");
MODULE_DESCRIPTION("Two Kernel Monte: Loads new Linux kernels from Linux.");
MODULE_PARM(run_setup, "0-1i");
MODULE_PARM_DESC(run_setup, "Set this to 1 to execute the setup portion of the Linux kernel image. (unsafe)");

#define MONTE_SETUP_BEGIN      0x90000
#define MONTE_CMDLINE_BEGIN    0x9C000
#define MONTE_INITRD_START     0x800000	/* Load the initrd at 8MB. */

#define BS_SIG_VAL     0xaa55
#define SETUP_SIG_VAL  "HdrS"
#define CMDLINE_MAGIC  0xA33F

#define DFL_SETUP_SECTS 4

struct kernel_setup_t {
    char __pad1[32];
    unsigned short cmd_magic;	/*  32: Command line magic 0xA33F */
    unsigned short cmd_offset;	/*  34: Command line offset from 0x90000 */
    char __pad2[461];
    unsigned char  setup_sects; /* 497: setup size in sectors (512) */
    unsigned short root_flags;	/* 498: 1 = ro ; 0 = rw */
    unsigned short kernel_para;	/* 500: kernel size in paragraphs (16) */
    unsigned short swap_dev;	/* 502: */
    unsigned short ram_size;	/* 504: */
    unsigned short vid_mode;	/* 506: */
    unsigned short root_dev;	/* 508: */
    unsigned short boot_flag;	/* 510: signature*/
    unsigned short jump;        /* 512: jump to startup code */
    char signature[4];          /* 514: "HdrS" */
    unsigned short version;     /* 518: header version */
    unsigned short x,y,z;       /* 520: LOADLIN hacks */
    unsigned short ver_offset;  /* 526: kernel version string */
    unsigned char loader;       /* 528: loader type */
    unsigned char flags;        /* 529: loader flags */
    unsigned short a;           /* 530: more LOADLIN hacks */
    unsigned long start;        /* 532: kernel start, filled in by loader */
    unsigned long ramdisk;      /* 536: RAM disk start address */
    unsigned long ramdisk_size; /* 540: RAM disk size */
    unsigned short b,c;         /* 544: bzImage hacks */
    unsigned short heap_end_ptr;/* 548: end of free area after setup code */
};

/* These are defined as macros like this because the address where
 * we're storing this stuff could change... (See memory management
 * below) This is safe w/o checking the result since we ask for this
 * page to write it before ever using this macro.  */
#define m_setup ((struct kernel_setup_t *)m_get_addr_phys(MONTE_SETUP_BEGIN))

/* User interface */
#define MONTE_MAGIC_1 0xdeaddead
#define MONTE_MAGIC_2 0xdeaddead
struct monte_param_t {
    char *command_line;

    int   kernel_fd;
    int   initrd_fd;

    /* XXX We should have an interface for copying a ramdisk or an
     * initrd from user space. */
};

/*--------------------------------------------------------------------
 * Monte memory management
 *--------------------------------------------------------------------
The problem:

 In order to boot a kernel we need to copy certain things to certain
 locations in memory.  It is likely that there could be kernel code or
 data already occupying these pages.  In the case of inital ramdisks,
 we don't necessarily need specific pages but allocating several
 contiguous megabytes of memory is not possible.

The Solution:

 We setup lists of pages and where they have to be moved to.  Then
 after we turn of paging during our switch to real mode, we run
 through this list and move all the blocks of memory to their final
 destinations.  (At that point we don't care about overwriting the
 kernel or our page table.)

 Setting up these lists is complicated by a few things.
  
  1 We don't know which pages we will want to use by the time we have
    to start allocating the first pages.

  2 We have to make sure that the movements can be performed safely in
    the sequence that they are given.

  3 The ASM code doing the movements must be as small as possible.

 Number 2 is the biggest issue.  It is resolved by looking for
 conflicts every time a new page is allocated in this system.

 The interface to this memory management code is m_get_addr() and
 m_get_addr_phys().

 void *m_get_addr(void *addr);
 void *m_get_addr_phys(unsigned long addr);

 These functions return the address of a page.  The data on that page
 will be moved to the address given by addr during the restart.  The
 addresses returned are in the kernel's virtual memory space.  The
 argument to m_get_addr() is an address in the kernel's virtual memory
 space.  The argument to m_get_addr_phys is a physical memory address.

 */


/* The list of movements to be done is stored in the following
 * structure.  If a page happens to fall on the address where it needs
 * to go, it still appears in this list.
 *
 * Pages used to store the list of pages (indirect pages) do not
 * appear in this list.
 *
 * The addresses in these structures are virtual addresses while these
 * structures are being setup.
 *
 * They are translated to physical addresses just before restart.
 */

struct monte_indirect_t {
    unsigned long dest;
    unsigned long src;
    unsigned long size;
    struct monte_indirect_t *next;
};
static struct monte_indirect_t *m_page_list=0;
static struct monte_indirect_t *m_unused_list=0;

/* Warning: This struct is referenced by ASM code below.  If you
 * re-arrange it, you'll have to fix the ASM as well. */

/*---------------------------------------------------------------
 *  monte_get_page - like get free page except that we're guaranteed to
 *                   not get a reserved page.  (unless we happen to find dest)
 *-------------------------------------------------------------*/
static
struct monte_indirect_t *m_find_dest(unsigned long addr) {
    struct monte_indirect_t *m = m_page_list;
    while (m) {
	if (m->dest == addr) { return m; }
	m = m->next;
    }
    return 0;
}

static
struct monte_indirect_t *m_find_src(unsigned long addr) {
    struct monte_indirect_t *m = m_page_list;
    while (m) {
	if (m->src == addr) return m;
	m = m->next;
    }
    return 0;
}

static
struct monte_indirect_t *m_find_ind(unsigned long addr) {
    /* Find an indirect block that's sitting on this address.  The
     * return value of this function probably isn't quite useful except as a yes/no flag */
    struct monte_indirect_t *m = m_page_list;
    while (m) {
	if ((((unsigned long)m)&PAGE_MASK) == addr) return m;
	m = m->next;
    }
    /* Check the unused list as well */
    m = m_unused_list;
    while (m) {
	if ((((unsigned long)m)&PAGE_MASK) == addr) return m;
	m = m->next;
    }
    return 0;
}

static
unsigned long m_get_new_page(void) {
    unsigned long pg, newpg;
    struct monte_indirect_t *m;

    pg = get_free_page(GFP_KERNEL);
    if (!pg) return 0;
    
    /* Check to see if somebody else wants this particular page */
    m = m_find_dest(pg);
    if (m) {
	memcpy((void *)pg, (void *)m->src, PAGE_SIZE);
	newpg = m->src;
	m->src = pg;		/* Update this page list entry. */
	pg = newpg;
    }
    return pg;
}

static
int m_add_page(unsigned long dest, unsigned long src) {
    int i;
    struct monte_indirect_t *m;

    if (!m_unused_list) {
	/* At this point, we've got a new page full of data but it's
           not on the list yet.  (We're about to add it.)  We need to
           allocate another page of indirect structs to do this since
           our unused list is empty.  If get_free_page happens to
           return dest, m_get_new_page() won't know to relocate this
           new page we've got unless it's on the list.  Therefore, it
           gets added in a temporary way here. */
	struct monte_indirect_t tmp;
	tmp.dest = dest;
	tmp.src  = src;
	tmp.size = PAGE_SIZE;
	tmp.next = m_page_list;
	m_page_list = &tmp;	/* Stick it on the list. */

	m = (struct monte_indirect_t *) m_get_new_page();
	if (!m) return -ENOMEM;

	m_page_list = tmp.next;	/* Remove it again. */

	/* Setup the next pointers on all these entries. */
	m[0].next = 0;
	for (i=1; i < PAGE_SIZE/sizeof(struct monte_indirect_t); i++)
	    m[i].next = &m[i-1];
	m_unused_list = &m[i-1];

	/* It's possible our new page got relocated. */
	dest = tmp.dest;
	src  = tmp.src;
    }
    
    /* Grab an item off the indirect list and throw it on our list of stuff */
    m = m_unused_list;
    m_unused_list = m_unused_list->next;

    m->dest = dest;
    m->src  = src;
    m->size = PAGE_SIZE;

    m->next = m_page_list;
    m_page_list = m;
    return 0;
}

static
int m_relocate_page(struct monte_indirect_t *m) {
    unsigned long pg;
    pg = m_get_new_page();
    if (pg == 0) return -ENOMEM;
    
    memcpy((void *) pg, (void *) m->src, PAGE_SIZE);
    m->src = pg;
    return 0;
}

static
void m_relocate_ind_page(unsigned long addr, struct monte_indirect_t **list, unsigned long pg) {
    struct monte_indirect_t *m, *last=0, *new_m;
    /* This is nasty.... we want to move all the indirect structs off
     * a particular page and onto the same spot on another page. */
    m = *list;
    while (m) {
	if ((((unsigned long)m)&PAGE_MASK) == addr) {
	    /* Ok we want to move this indirect entry */
	    new_m = (void *)(pg + (((unsigned long)m)&~PAGE_MASK));
	    *new_m = *m;

	    /* Keep those next pointers up to date. */
	    if (last)
		last->next = new_m;
	    else
		*list = new_m;
	}
	last = m;
	m = m->next;
    }
}

static
int m_relocate_ind(unsigned long addr) {
    unsigned long pg;
    struct monte_indirect_t *m, *last, *new_m;

    /* First off, grab a new page to move them to. */
    pg = m_get_new_page();
    if (pg == 0) return -ENOMEM;

    /* Go through our page list transplanting entries */
    m_relocate_ind_page(addr, &m_page_list, pg);
    m_relocate_ind_page(addr, &m_unused_list, pg);
    return 0;
}
	

static
unsigned long m_get_page(unsigned long addr) {
    int err;
    unsigned long pg;
    struct monte_indirect_t *m;

    /* First check if we have this page on our list somewhere */
    m = m_find_dest(addr);
    if (m) return m->src;

    /* This page isn't on our lists.  It's possible that we've already
     * used the page we want as an indirect page or as a source page
     * for another destination. */

    /* Check for indirect pages sitting on this address.  If so, move the indirect page. */
    m = m_find_ind(addr);
    if (m) {
	if ((err = m_relocate_ind(addr))) return err;
	m_add_page(addr,addr);
	return addr;
    }

    /* See if another page's source page is sitting on the dest page
     * we want to be in.  If so, relocate that page and stick
     * ourselves in that page. */
    m = m_find_src(addr);
    if (m) {
	/* Another data page is sitting on the spot we want to be.  We'll have to move it. */
	if ((err = m_relocate_page(m))) return err;
	m_add_page(addr,addr);
	return addr;
    }

    /* Ok, finally, the simple case. */
    pg = m_get_new_page();
    m_add_page(addr, pg);
    return pg;
}

static
void *m_get_addr(void *addr) {
    unsigned long pg;
    pg = m_get_page(((unsigned long)addr)&PAGE_MASK);
    if (!pg) return 0;
    return (void *)(pg + (((unsigned long)addr)&~PAGE_MASK));
}

static
void *m_get_addr_phys(unsigned long addr) {
    if (addr >= __pa(high_memory)) return 0;
    return m_get_addr(__va(addr));
}


static
void m_make_phys(void) {
    struct monte_indirect_t *m = m_page_list, *next_m;

    /* translate all the address to physical addresses for use by the
     * boot loader code. */
    while (m) {
	next_m = m->next;
	m->dest = __pa(m->dest);
	m->src  = __pa(m->src);
	if (m->next) m->next = (struct monte_indirect_t *) __pa(m->next);
	m = next_m;
    }
}

static
void m_free_all_ind(struct monte_indirect_t **list) {
    unsigned long *ctr;
    struct monte_indirect_t *m, *next_m;

    /* There are 256 entries per page.  Use the dest field of the
     * first entry on a page as a counter.  Start deleting entries
     * from the lists.  When the counter on the page reaches zero,
     * free the indirect page. */

    m = *list;
    *list = 0;
    while (m) {
	next_m = m->next;
	/* Update the page counter */
	ctr = (unsigned long *)(((unsigned long)m)&PAGE_MASK);
	(*ctr)--;
	if (*ctr == 0) {
	    printk("m_free_all_ind: freeing %08lx\n", (long)ctr);
	    free_page((unsigned long)ctr);
	}
	m = next_m;
    }
}

static
void m_free_all(void) {
    struct monte_indirect_t *m;

    /* Pass 1: free all data pages and initialize the counters for pass 2 */
    m = m_page_list;
    while (m) {
	free_page(m->src);
	m->dest = PAGE_SIZE/sizeof(struct monte_indirect_t); /* init for pass2 */
	m = m->next;
    }

    /* Pass 2: free the indirect pages */
    m_free_all_ind(&m_page_list);
    m_free_all_ind(&m_unused_list);
}


/*--------------------------------------------------------------------
 * Kernel Image loading
 *------------------------------------------------------------------*/

/* Tidbits to allow easy file reads from kernel space */
extern void * sys_call_table[];
#define k_sys_call3(rt,sys,t1,arg1,t2,arg2,t3,arg3) \
            (((rt (*)(t1,t2,t3)) \
              sys_call_table[(sys)])((arg1),(arg2),(arg3)))
#define k_sys_read(fd,buf,count) k_sys_call3(int,__NR_read,int,(fd),char*,(buf),size_t,(count))

static
ssize_t k_read(int fd, void *buffer, size_t len) {
    int err;
    mm_segment_t oldfs;
    oldfs = get_fs(); set_fs(KERNEL_DS);
    err = k_sys_read(fd, buffer, len);
    set_fs(oldfs);
    return err;
}

static
ssize_t k_read_phys(int fd, unsigned long buffer, size_t bytes) {
    int len, r, bytes_read = 0;
    void *dest;
    
    /* We have to be careful not to read across page boundaries. */

    while (bytes_read < bytes) {
	dest = m_get_addr_phys(buffer);
	
	/* Read up until the end of the page only */
	len = PAGE_SIZE - (buffer & ~PAGE_MASK);
	if (len > bytes-bytes_read) len = bytes-bytes_read;

	r = k_read(fd, dest, len);
	if (r == 0) return bytes_read;
	if (r < 0)  return bytes_read ? bytes_read : r;

	bytes_read += r;
	buffer += r;
    }
    return bytes_read;
}

static
int monte_load_kernel(int fd) {
    int r, i, len;
    void *ptr;
    unsigned long phys_ptr;

    /* Load the kernel's boot sector */
    r = k_read_phys(fd, MONTE_SETUP_BEGIN, 512);
    if (r < 0)    return r;
    if (r != 512) return -EIO;

    if (m_setup->boot_flag != BS_SIG_VAL) { 
	printk("monte: Boot signature not found.\n");
	return -EINVAL;
    }

    printk("monte: Loading %d sectors of setup code.\n", (int) m_setup->setup_sects);

    /* Load the kernel's setup code.. */
    phys_ptr = MONTE_SETUP_BEGIN+512;
    for (i=0; i < (int) m_setup->setup_sects; i++) {
	r = k_read_phys(fd, phys_ptr, 512);
	if (r != 512) {
	    if (r < 0) return r;
	    printk("monte: Short read on setup data.\n");
	    return -EIO;
	}
	phys_ptr += 512;
    }

    /* Check for that setup signature. */
    if (strncmp(m_setup->signature, SETUP_SIG_VAL, strlen(SETUP_SIG_VAL)) != 0) {
	printk("monte: Kernel image setup signature not found.\n");
	return -EINVAL;
    }

    /* This kernel size is rounded up to the next whole paragraph. */
    len = m_setup->kernel_para*16;
    printk("monte: Loading kernel data at 0x%lx\n",
	   m_setup->start);
    r = k_read_phys(fd, m_setup->start, len);
    if (r < 0) return r;
    if (len - r >= 16) {
	printk("monte: Short read on kernel data.\n");
	return -EIO;
    }
    printk("monte: Loaded %ld bytes of kernel data.\n", r);

    if (!run_setup) {
	/* The setup portion normally sets up a lot of data for the
	 * kernel.  Most of this stuff is stored in the boot sector.
	 * Since we're not running setup.  We should preserve the copy
	 * of the boot sector from the previous boot.  (This presumes
	 * that the format of this data won't change between kernel
	 * revisions.  The kernel isn't changing THAT fast is it?)
	 */
	memcpy(m_setup, __va(MONTE_SETUP_BEGIN), 512);
    }
    return 0;
}

static
int monte_load_initrd(int fd) {
    int r;
    unsigned long phys_ptr;
    phys_ptr = m_setup->ramdisk;
    while (1) {
	/* k_read_phys will fail if we run over the end of memory. */
	r = k_read_phys(fd, phys_ptr, PAGE_SIZE);
	if (r < 0) return r;
	if (r == 0) {
	    m_setup->ramdisk_size = phys_ptr - m_setup->ramdisk;
	    printk("monte: Loaded %ld bytes of initial ramdisk data.\n",
		   m_setup->ramdisk_size);
	    return 0;
	}

	phys_ptr += r;
    }
}

/*--------------------------------------------------------------------
 *  Syscall interface
 *------------------------------------------------------------------*/
static int monte_restart(void);
int (*real_reboot)(int, int, int, void *);
asmlinkage int sys_monte(int magic1, int magic2, int cmd, void *arg) {
    int err;
    char *cmdline, *ptr;
    struct monte_param_t   param;
    struct setup_header_t *hdr;

    MOD_INC_USE_COUNT;
    if (magic1 != MONTE_MAGIC_1 || magic2 != MONTE_MAGIC_2) {
	err = real_reboot(magic1, magic2, cmd, arg);
	MOD_DEC_USE_COUNT;
	return err;
    }

    /* XXX Should really protect entry here with a semaphore.. */

    m_page_list   = 0;
    m_unused_list = 0;

    if (!capable(CAP_SYS_BOOT)) {
	err = -EPERM;
	goto sys_monte_bail_out;
    }

    /* Get all the arguments from the user */
    if (copy_from_user(&param, arg, sizeof(param))) {
	err = -EPERM;
	goto sys_monte_bail_out;
    }

    cmdline = (char *) get_free_page(GFP_KERNEL);
    if (!cmdline) {
	err = -ENOMEM;
	goto sys_monte_bail_out;
    }
    err = strncpy_from_user(cmdline, param.command_line, PAGE_SIZE);
    if (err < 0) {
	err = -EFAULT;
	goto sys_monte_bail_out_fp;
    }
    if (err >= PAGE_SIZE) {
	err = -ENAMETOOLONG;
	goto sys_monte_bail_out_fp;
    }

    printk("monte: command line: %s\n", param.command_line);

    err = monte_load_kernel(param.kernel_fd);
    if (err) goto sys_monte_bail_out_fp;

    /* Load the initrd */
    if (param.initrd_fd != -1) {
	/* We have to decide where to drop the initial ram disk. It's got
	 * to be out of the way of the kernel even when uncompressed. */
	m_setup->ramdisk = MONTE_INITRD_START;
	err = monte_load_initrd(param.initrd_fd);
	if (err) goto sys_monte_bail_out_fp;
    } else {
	/* No initial ramdisk */
	m_setup->ramdisk      = 0;
	m_setup->ramdisk_size = 0;
    }

    m_setup->loader = 0x50;	/* Set the loader type. */

    /* Setup the command line */
    m_setup->cmd_magic  = CMDLINE_MAGIC;
    m_setup->cmd_offset = MONTE_CMDLINE_BEGIN - MONTE_SETUP_BEGIN;
    ptr = m_get_addr_phys(MONTE_CMDLINE_BEGIN);
    if (!ptr) {
	err = -ENOMEM;
	goto sys_monte_bail_out_fp;
    }
    strcpy(ptr, cmdline);
	    
    /* Restart with a new kernel (from memory) */
    printk("monte: restarting system...\n");
    err = monte_restart();

    /* If we come back some error ocurred... drop through to the error
     * cleanup */

 sys_monte_bail_out_fp:
    free_page((unsigned long)cmdline);
 sys_monte_bail_out:
    m_free_all();
    MOD_DEC_USE_COUNT;
    return err;
}


/*-------------------------------------------------------------------------
 * Machine restart code
 *-----------------------------------------------------------------------*/
static unsigned long long
real_mode_gdt_entries [] = {
        0x0000000000000000ULL,  /* 00h: Null descriptor */
	0x0000000000000000ULL,  /* 08h: Unused... */
	0x00cf9a000000ffffULL,	/* 10h: 32-bit 4GB code at 0x00000000 */
	0x00cf92000000ffffULL,	/* 18h: 32-bit 4GB data at 0x00000000 */
        0x00009a000000ffffULL,  /* 20h: 16-bit 64k code at 0x00000000 */
        0x000092000000ffffULL	/* 28h: 16-bit 64k data at 0x00000000 */
};

static struct {
        unsigned short       size __attribute__ ((packed));
        unsigned long long * base __attribute__ ((packed));
}
real_mode_gdt = { sizeof (real_mode_gdt_entries)-1, 0 },
real_mode_idt = { 0x3ff, 0 };

/* The address arguments to this function are PHYSICAL ADDRESSES */ 
static void real_mode_switch(struct monte_indirect_t *mem_list,
			     unsigned long flags) {
  __asm__ __volatile__
      (/* Grab the args for our memcpy off the stack now while we still
	* have page tables. */
       "    movl %0, %%ebx           \n" /* Save our argument in ebx */
       "    movl %1, %%edx           \n" /* Store our flags in edx */

       /* Turn off paging, leave protection turned on. */
       "    movl %%cr0, %%eax        \n" /* Turn off paging (bit 31 in CR0) */
       "    andl $0x7FFFFFFF, %%eax  \n"
       "    movl %%eax, %%cr0        \n"
       "    xorl %%eax, %%eax        \n" /* Flush the TLB (write 0 to CR3) */
       "    movl %%eax, %%cr3        \n"

       /* Memory setup code */
       /* We need to memcpy all the bits of code and data reload to
	* the right places now We can do this safely now that we don't
	* have to worry about overwriting our own page tables.  */
       "    mov  $0x18, %%ax         \n" /* 32-bit, 4GB data from GDT */
       "    mov  %%ax, %%ds          \n"
       "    mov  %%ax, %%es          \n"

       "    cld                      \n"
       "__mem_setup_loop:            \n"
       "    orl  %%ebx, %%ebx        \n" /* ebx = addr of chunk descriptor */
       "    jz   __mem_finished      \n"
       "    movl  (%%ebx), %%edi     \n" /* Destination */
       "    movl 4(%%ebx), %%esi     \n" /* Source */
       "    movl 8(%%ebx), %%ecx     \n" /* Size */
       "    cmpl %%edi, %%esi        \n" /* skip if dest == src  */
       "    je   __mem_skip          \n"
       "    or   %%ecx, %%ecx        \n" /* skip if size == 0 */
       "    jz   __mem_skip          \n"
       "    rep; movsb               \n" /* Do it. */
       "__mem_skip:                  \n"
       "    movl 12(%%ebx), %%ebx    \n"
       "    jmp  __mem_setup_loop    \n"
       "__mem_finished:              \n"

       /* Option here... we can skip the "setup" portion of the kernel
	* or not.
	* bit 0 of EDX:  0 = skip it, 1 = run it (unsafe)
	*/
       "    andl $0x1, %%edx         \n"
       "    jnz  __rms_run_setup     \n"
       "        xor  %%bx, %%bx      \n" /* bx==0 means boot */
       "        ljmp $0x10, $0x100000\n"

       "__rms_run_setup:             \n"
       /* Now that our memcpy is done we can get to 16 bit code
	* segment.  This configures CS properly for real mode. */
       "    ljmp $0x20, $0x1000-(real_mode_switch_end - __rms_16bit) \n"
       "__rms_16bit:                 \n"
       ".code16                      \n" /* 16 bit code from here on... */

       /* Load the segment registers w/ properly configured segment
	* descriptors.  They will retain these configurations (limits,
	* writability, etc.) once protected mode is turned off. */
       "    mov  $0x28, %%ax         \n"
       "    mov  %%ax, %%ds          \n"
       "    mov  %%ax, %%es          \n"
       "    mov  %%ax, %%fs          \n"
       "    mov  %%ax, %%gs          \n"
       "    mov  %%ax, %%ss          \n"

       /* Turn off protection (bit 0 in CR0) */
       "    movl %%cr0, %%eax        \n"
       "    andl $0xFFFFFFFE, %%eax  \n"
       "    movl %%eax, %%cr0        \n"

       /* Now really going into real mode */
       "    ljmp $0, $0x1000-(real_mode_switch_end - __rms_real) \n"
       "__rms_real:                  \n"

       /* Setup a stack */
       "    mov  $0x9000, %%ax       \n"
       "    mov  %%ax, %%ss          \n"
       "    mov  $0xAFFE, %%ax       \n"
       "    mov  %%ax, %%sp          \n"

       /* Dump zeros in the other segregs */
       "    xor  %%ax, %%ax          \n"
       "    mov  %%ax, %%ds          \n"
       "    mov  %%ax, %%es          \n"
       "    mov  %%ax, %%fs          \n"
       "    mov  %%ax, %%gs          \n"

       "    sti                      \n" /* Enable interrupts */

       "    ljmp $0x9020, $0x0000    \n"
#if 0
       /* Debugging tools... */
       "    .byte 0xcc               \n" /* XXX DEBUG triple fault XXXX */
       "    .byte 0xeb, 0xfe         \n" /* XXX DEBUG wedge XXXXX */
#endif
       ".code32                      \n" /* Restore mode for rest of file */
       : : "m" (mem_list), "m" (flags));
}
__asm__ (".text\n""real_mode_switch_end:\n");
extern char real_mode_switch_end[];

#ifdef CONFIG_PCI
static
void monte_pci_disable_bus(struct pci_bus *bus) {
    struct pci_dev *dev;
    u16 cmd;
    struct list_head *l;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,1)
    for (dev=bus->devices; dev!=NULL; dev=dev->next) {
#else
    for (l=bus->devices.next; l != &bus->devices; l=l->next) {
	dev = pci_dev_b(l);
#endif
	pcibios_read_config_word(dev->bus->number, dev->devfn, PCI_COMMAND, &cmd);
	cmd &= ~PCI_COMMAND_MASTER;
	pcibios_write_config_word(dev->bus->number, dev->devfn, PCI_COMMAND, cmd);
    }
}

static
void monte_pci_disable(void) {
    struct pci_bus *bus;
    struct list_head *l;
    /* Turn off PCI bus masters to keep them from scribbling on our
     * memory later on. */
    if (pcibios_present()) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,1)
	for (bus=&pci_root; bus != NULL; bus=bus->next)
	    monte_pci_disable_bus(bus);
#else
	for (l=pci_root_buses.next; l != &pci_root_buses; l=l->next)
	    monte_pci_disable_bus(pci_bus_b(l));
#endif
    }
}
#endif /* CONFIG_PCI */

static
int monte_restart(void) {
    int i;
    void * ptr;
    pgd_t *pgd;
    pmd_t *pmd;
    pte_t *pte;

    cli();

    /* Step 1: get an identity mapped page.  The code that will turn
     * off paging later needs to run out of an identity mapped page.
     * For simplicity we'll use page zero.  This page is normally not
     * mapped at all.  This code might look pretty generic but it's
     * really highly x86 specific.
     *
     * We'll operate on init_mm's page table. */
    pgd = pgd_offset(&init_mm, 0); /* pgd_none() always 0 on x86... */
    pmd = pmd_offset(pgd, 0);
    if (pmd_none(*pmd)) {
	/* If there' is no page table under this middle entry, make
	 * one.  We do it by hand since the symbols to do it are not
	 * exported.  */
	pte = (pte_t *) get_free_page(GFP_KERNEL);
	if (!pte) {
	    sti();
	    printk("monte: out of memory.\n");
	    return -ENOMEM;
	}
/* !!!!!! The point of no return is here. !!!!!! */
	memset((void *)pte, 0, PAGE_SIZE); /* New page table... */
	pmd_val(*pmd) = __pa(pte) | _PAGE_RW | _PAGE_PRESENT; /* Poke it into the page dir. */
    }
    /* Map page zero at physical address 0x0 and mark it as present. */
    pte = pte_offset(pmd, 0);
    pte_val(*pte) = 0x0 | _PAGE_RW | _PAGE_PRESENT;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,1)
    SET_PAGE_DIR(current, init_mm.pgd);	/* Install the page table */
#else
	asm volatile("movl %0,%%cr3": :"r" (__pa(init_mm.pgd)));
#endif

    /* Step 2: Copy the real_mode_switch code and our GDT into the
     * page we just setup.  We stick this stuff right at the end to
     * avoid clobbering BIOS variables and stuff in the beginning. */
#define RMS_SIZE ((unsigned long)real_mode_switch_end - (unsigned long) real_mode_switch)
    /* Copy our GDT down there. */
    ptr = (void *) ((0x1000 - RMS_SIZE - sizeof(real_mode_gdt_entries)) & ~7);
    memcpy(ptr, real_mode_gdt_entries, sizeof(real_mode_gdt_entries));
    real_mode_gdt.base = ptr;

    /* Copy the real_mode_switch function down there. */
    memcpy((void *)(0x1000-RMS_SIZE), real_mode_switch, RMS_SIZE);

#ifdef CONFIG_PCI
    monte_pci_disable();	/* Turn off PCI bus masters */
#endif
    m_make_phys();		/* Prep our memory to-do list. */
    /*m_make_phys_sanity();*/

    /* Ok, now the real monkey business begins.... Please keep hands
     * and feet inside the memory space and remain seated until the
     * ride comes to a complete stop. */
    __asm__ __volatile__
	(/* Install the IDT and GDT we copied to page zero. */
	 "lidt  %0             \n"
	 "lgdt  %1             \n"
	 
	 /* Function call with args... sort of */
	 "pushl %3             \n"
	 "pushl %2             \n" /* Push args on the stack */
	 "sub $4, %%esp        \n" /* bogo return address */
	 "ljmp $0x10, $0x1000-(real_mode_switch_end-real_mode_switch) \n" : :
	 "m" (real_mode_idt), "m" (real_mode_gdt), 
	 "r" (__pa(m_page_list)), "r"(run_setup)
	 : "memory");
    /* NOT REACHED */
    while(1);			/* Shut up gcc. */
}


/*--------------------------------------------------------------------
 *  Module mechanics
 *------------------------------------------------------------------*/
extern void *sys_call_table[];
int init_module(void) {
    printk("monte: 2 Kernel Monte Version %s\n"
	   "monte: Erik Arjan Hendriks <hendriks@scyld.com>\n",
	   PACKAGE_VERSION);

    real_reboot = sys_call_table[__NR_reboot];
    sys_call_table[__NR_reboot] = sys_monte;
    return 0;
}

void cleanup_module(void) {
    sys_call_table[__NR_reboot] = real_reboot;
}

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
